@@ -38,7 +38,7 @@ CMD_FUNC(cmd_mlock);
 void _do_mode(Channel *channel, Client *client, MessageTag *recv_mtags, int parc, const char *parv[], time_t sendts, int samode);
 MultiLineMode *_set_mode(Channel *channel, Client *client, int parc, const char *parv[], u_int *pcount,
                        char pvar[MAXMODEPARAMS][MODEBUFLEN + 3]);
-void _set_channel_mode(Channel *channel, const char *modes, const char *parameters);
+void _set_channel_mode(Channel *channel, MessageTag *mtags, const char *modes, const char *parameters);
 CMD_FUNC(_cmd_umode);
 
 /* local: */
@@ -312,6 +312,22 @@ void _do_mode(Channel *channel, Client *client, MessageTag *recv_mtags, int parc
 		MessageTag *mtags = NULL;
 		int should_destroy = 0;
 
+		if (IsUser(orig_client) && samode && MyUser(orig_client))
+		{
+			if (!sajoinmode)
+			{
+				char buf[512];
+				snprintf(buf, sizeof(buf), "%s%s%s", modebuf, *parabuf ? " " : "", parabuf);
+				unreal_log(ULOG_INFO, "samode", "SAMODE_COMMAND", orig_client,
+					   "Client $client used SAMODE $channel ($mode)",
+					   log_data_channel("channel", channel),
+					   log_data_string("mode", buf));
+			}
+
+			client = &me;
+			sendts = 0;
+		}
+
 		if (m->numlines == 1)
 		{
 			/* Single mode lines are easy: retain original msgid etc */
@@ -350,22 +366,6 @@ void _do_mode(Channel *channel, Client *client, MessageTag *recv_mtags, int parc
 		}
 #endif
 
-		if (IsUser(orig_client) && samode && MyUser(orig_client))
-		{
-			if (!sajoinmode)
-			{
-				char buf[512];
-				snprintf(buf, sizeof(buf), "%s%s%s", modebuf, *parabuf ? " " : "", parabuf);
-				unreal_log(ULOG_INFO, "samode", "SAMODE_COMMAND", orig_client,
-					   "Client $client used SAMODE $channel ($mode)",
-					   log_data_channel("channel", channel),
-					   log_data_string("mode", buf));
-			}
-
-			client = &me;
-			sendts = 0;
-		}
-
 		sendto_channel(channel, client, NULL, 0, 0, SEND_LOCAL, mtags,
 			       ":%s MODE %s %s %s",
 			       client->name, channel->name, modebuf, parabuf);
@@ -593,10 +593,10 @@ const char *mode_ban_handler(Client *client, Channel *channel, const char *param
 		}
 	}
 
-	if ( (what == MODE_ADD && add_listmode(banlist, client, channel, tmpstr)) ||
+	if ( (what == MODE_ADD && (add_listmode(banlist, client, channel, tmpstr) != 1)) ||
 	     (what == MODE_DEL && del_listmode(banlist, channel, tmpstr)))
 	{
-		return NULL;	/* already exists */
+		return NULL;	/* ban to be added already exists, or ban to be deleted does not exist */
 	}
 
 	return tmpstr;
@@ -1087,6 +1087,7 @@ CMD_FUNC(_cmd_umode)
 	int what;
 	long oldumodes = 0;
 	char oldsnomask[64];
+	const char *restrict_usermodes = NULL;
 	/* (small note: keep 'what' as an int. -- Syzop). */
 	short rpterror = 0, umode_restrict_err = 0, chk_restrict = 0, modex_err = 0;
 
@@ -1125,8 +1126,8 @@ CMD_FUNC(_cmd_umode)
 
 	oldumodes = client->umodes;
 
-	if (RESTRICT_USERMODES && MyUser(client) && !ValidatePermissionsForPath("immune:restrict-usermodes",client,NULL,NULL,NULL))
-		chk_restrict = 1;
+	if (MyUser(client) && !ValidatePermissionsForPath("immune:restrict-usermodes",client,NULL,NULL,NULL))
+		restrict_usermodes = get_setting_for_user_string(client, SET_RESTRICT_USERMODES);
 
 	if (client->user->snomask)
 		strlcpy(oldsnomask, client->user->snomask, sizeof(oldsnomask));
@@ -1136,12 +1137,12 @@ CMD_FUNC(_cmd_umode)
 	 */
 	for (m = parv[2]; *m; m++)
 	{
-		if (chk_restrict && strchr(RESTRICT_USERMODES, *m))
+		if (restrict_usermodes && strchr(restrict_usermodes, *m))
 		{
 			if (!umode_restrict_err)
 			{
 				sendnotice(client, "Setting/removing of usermode(s) '%s' has been disabled.",
-					RESTRICT_USERMODES);
+					restrict_usermodes);
 				umode_restrict_err = 1;
 			}
 			continue;
@@ -1545,7 +1546,7 @@ int list_mode_request(Client *client, Channel *channel, const char *req)
 	return 1; /* handled */
 }
 
-void _set_channel_mode(Channel *channel, const char *modes, const char *parameters)
+void _set_channel_mode(Channel *channel, MessageTag *mtags, const char *modes, const char *parameters)
 {
 	char buf[512];
 	char *p, *param;
@@ -1561,7 +1562,7 @@ void _set_channel_mode(Channel *channel, const char *modes, const char *paramete
 	myparv[myparc] = NULL;
 
 	SetULine(&me); // hack for crash.. set ulined so no access checks.
-	do_mode(channel, &me, NULL, myparc, (const char **)myparv, 0, 0);
+	do_mode(channel, &me, mtags, myparc, (const char **)myparv, 0, 0);
 	ClearULine(&me); // and clear it again..
 
 	for (i = 0; i < myparc; i++)
